/*
 * Copyright 2000 Hans Reiser
 */

#include "fsck.h"
#include <limits.h>


//
// set up start journal block and journal size
// make journal unreplayable by kernel replay routine
//
void reset_journal (struct super_block * s)
{
    int i ;
    struct buffer_head *bh ;
    int done = 0;
    int len;
    int start;

    // FIXME: we leave journal start and journal size unchanged as we
    // hope they are correct and actually we have to way to guess
    // about them anything else
    start = rs_journal_start (s->u.reiserfs_sb.s_rs);

    /* journal size */
    len = rs_journal_size (s->u.reiserfs_sb.s_rs) + 1;

    printf ("Resetting journal - %d blocks starting from %d\n",
	    len, start); fflush (stdout);

    for (i = 0 ; i < len ; i++) {
	print_how_far (&done, len);
        bh = getblk (s->s_dev, start + i, s->s_blocksize);
	memset(bh->b_data, 0, s->s_blocksize) ;
	mark_buffer_dirty(bh,0) ;
//	mark_buffer_uptodate(bh,0) ;
        mark_buffer_uptodate(bh,1) ;
	if (bwrite (bh) == -1)
	    die ("reset_journal: unable to read %lu block on device 0x%x\n",
			      start + i, s->s_dev);
	brelse(bh) ;
    }
    printf ("\n"); fflush (stdout);
    
#if 0 /* need better way to make journal unreplayable */


    /* have journal_read to replay nothing: look for first non-desc
       block and set j_first_unflushed_offset to it */
    {   
	int offset;
	struct buffer_head * bh, *jh_bh;
	struct reiserfs_journal_header * j_head;
	struct reiserfs_journal_desc * desc;


	jh_bh = bread (s->s_dev, s->u.reiserfs_sb.s_rs->s_journal_block + s->u.reiserfs_sb.s_rs->s_orig_journal_size,
		       s->s_blocksize);
        if (jh_bh == 0)
            reiserfs_panic(s, "reset_journal: unable to read %lu block on device 0x%x\n",
                s->u.reiserfs_sb.s_rs->s_journal_block + s->u.reiserfs_sb.s_rs->s_orig_journal_size, s->s_dev);
	j_head = (struct reiserfs_journal_header *)(jh_bh->b_data);

	for (offset = 0; offset < s->u.reiserfs_sb.s_rs->s_orig_journal_size; offset ++) {
	    bh = bread (s->s_dev, s->u.reiserfs_sb.s_rs->s_journal_block + offset, s->s_blocksize);
            if (bh == 0)
                reiserfs_panic(s, "reset_journal: unable to read %lu block on device 0x%x\n",
                    s->u.reiserfs_sb.s_rs->s_journal_block + offset, s->s_dev);
	    desc = (struct reiserfs_journal_desc *)((bh)->b_data);
	    if (memcmp(desc->j_magic, JOURNAL_DESC_MAGIC, 8)) {
		/* not desc block found */
		j_head->j_first_unflushed_offset = offset;
		brelse (bh);
		break;
	    }
	    brelse (bh);
	}

	mark_buffer_uptodate (jh_bh, 1);
	mark_buffer_dirty (jh_bh, 1);
	bwrite (jh_bh);
	brelse (jh_bh);
    }
#endif
}

//
// end of stolen from ./fs/reiserfs/journal.c
//


#define bh_desc(bh) ((struct reiserfs_journal_desc *)((bh)->b_data))
#define bh_commit(bh) ((struct reiserfs_journal_commit *)((bh)->b_data))





static int desc_block (struct buffer_head * bh)
{
    struct reiserfs_journal_desc * desc = (struct reiserfs_journal_desc *)bh->b_data;
    if (!memcmp(desc->j_magic, JOURNAL_DESC_MAGIC, 8))
	return 1;
    return 0;
}

static int next_expected_desc (struct super_block * s, struct buffer_head * d_bh)
{
    int offset;
    struct reiserfs_journal_desc * desc;

    desc = (struct reiserfs_journal_desc *)d_bh->b_data;
    offset = d_bh->b_blocknr - SB_JOURNAL_BLOCK (s);
    return SB_JOURNAL_BLOCK (s) + ((offset + desc->j_len + 1 + 1) % JOURNAL_BLOCK_COUNT);
}


static int is_valid_transaction (struct super_block * s, struct buffer_head * d_bh)
{
    struct buffer_head * c_bh;
    int offset;
    struct reiserfs_journal_desc *desc  = (struct reiserfs_journal_desc *)d_bh->b_data;
    struct reiserfs_journal_commit *commit ;
    __u32 block, start_block;


    offset = d_bh->b_blocknr - SB_JOURNAL_BLOCK (s);
    
    /* ok, we have a journal description block, lets see if the transaction was valid */
    block = next_expected_desc (s, d_bh) - 1;
    start_block  = d_bh->b_blocknr;
    while(!(c_bh = bread (s->s_dev, block, s->s_blocksize))){
        if (++block == SB_JOURNAL_BLOCK (s) + JOURNAL_BLOCK_COUNT)
            block = SB_JOURNAL_BLOCK (s);
        if (block == start_block)
            return 0;
    }

    commit = (struct reiserfs_journal_commit *)c_bh->b_data ;
    if (does_desc_match_commit (s, desc, commit)) {
      //    if (journal_compare_desc_commit (s, desc, commit)) {
/*	printf ("desc and commit block do not match\n");*/
	brelse (c_bh) ;
	return 0;
    }
    brelse (c_bh);
    return 1;
}


static int next_desc (struct super_block * s, int this)
{
    struct buffer_head * bh;
    int retval;
    __u32 block;

    this = (this - SB_JOURNAL_BLOCK (s)) % JOURNAL_BLOCK_COUNT + SB_JOURNAL_BLOCK (s);
    block = this + 1;
    do {
        if (block == this)
            return 0;
        while(!(bh = bread (s->s_dev, block, s->s_blocksize)))
        {
            if (++block == SB_JOURNAL_BLOCK (s) + JOURNAL_BLOCK_COUNT)
                block = SB_JOURNAL_BLOCK (s);
            if (block == this)
                return 0;
        }

        if (!desc_block (bh)) {
            brelse (bh);
	        continue;
    	}

        retval = block;
        brelse (bh);

        break;
    } while (1);

    return retval;
}


void replay_all (struct super_block * s)
{
    int first_journal_block = SB_JOURNAL_BLOCK (s);
    int journal_size = rs_journal_size (s->u.reiserfs_sb.s_rs);
    struct buffer_head * d_bh, * c_bh, *jh_bh;
    struct reiserfs_journal_desc *desc ;
    struct reiserfs_journal_commit *commit ;
    int i;
    int the_most_old_transaction = INT_MAX;
    int the_most_young_transaction = 0;
    int valid_transactions = 0;
    int last_replayed;
    int start_replay = 0;
    struct reiserfs_journal_header * j_head;
    __u32 block;

    if (opt_journal_replay == NO_REPLAY)
        return;

    if (opt_journal_replay == REPLAY_DEFAULT)
    {
        block = first_journal_block + journal_size;
        while(!(jh_bh = bread (s->s_dev, block, s->s_blocksize))){
            if (--block == first_journal_block){
            /* we have only one good block */
                reiserfs_warning("replay_all: all journal blocks are bad - do not use journal anymore\n");
                return;
            }
        }

        j_head = (struct reiserfs_journal_header *)(jh_bh->b_data);
        start_replay = first_journal_block + j_head->j_first_unflushed_offset;
        if ((start_replay < first_journal_block) || (start_replay >= first_journal_block + journal_size))
        {
            opt_journal_replay == REPLAY_ALL;
            start_replay = 0;
        }else{
            block = start_replay;
            while(!(d_bh = bread (s->s_dev, block, s->s_blocksize))){
                if (++block == first_journal_block + journal_size)
                    block = first_journal_block;
                if (block == start_replay){
                    opt_journal_replay == REPLAY_ALL;
                    start_replay = 0;
                }
            }
            if (!is_valid_transaction (s, d_bh))
            {
                opt_journal_replay == REPLAY_ALL;
                start_replay = 0;
            }else{
                the_most_old_transaction = ((struct reiserfs_journal_desc *)d_bh->b_data)->j_trans_id;
                the_most_young_transaction = the_most_old_transaction;
            }
        }
    }

    /* look for oldest valid transaction for REPLAY_ALL or just get valid_transactions for REPLAY_DEFAULT */
    printf ("Looking for the oldest transaction to start with %4d", valid_transactions);
    for (i = first_journal_block; i < first_journal_block + journal_size; i ++) {
        while(!(d_bh = bread (s->s_dev, i, s->s_blocksize))){
            if (++i == first_journal_block + journal_size)
                goto end_of_for;
        }

        if (desc_block (d_bh)) {
            desc = (struct reiserfs_journal_desc *)d_bh->b_data;
            /*printf ("block %ld is desc block of the transaction (trans_id %ld, len %ld, mount_id %ld) - ",
                d_bh->b_blocknr, desc->j_trans_id, desc->j_len, desc->j_mount_id);*/
            if (!is_valid_transaction (s, d_bh)) {
                i += desc->j_len + 1;
                brelse (d_bh);
                continue;
            }

            if (the_most_young_transaction < desc->j_trans_id)
                the_most_young_transaction = desc->j_trans_id;

            if ((opt_journal_replay == REPLAY_ALL) && (the_most_old_transaction > desc->j_trans_id))
            {
                the_most_old_transaction = desc->j_trans_id;
                start_replay = d_bh->b_blocknr;
            }

            if ((opt_journal_replay == REPLAY_ALL) || (the_most_old_transaction <= desc->j_trans_id))
            {
                valid_transactions ++;
               /*printf ("good\n");*/
                printf ("\b\b\b\b    \b\b\b\b%4d", valid_transactions); fflush (stdout);
            }
            i += desc->j_len + 1;
        }
        brelse (d_bh);
        continue;
    }
end_of_for:

    printf ("\b\b\b\b     \b\b\b\bok\n"
        "%d valid trans found. Will replay from %d to %d\n", valid_transactions,
        the_most_old_transaction, the_most_young_transaction);


    printf ("Replaying transaction..%4d left..\b\b\b\b\b\b\b", valid_transactions);

    /* replay all valid transaction */
    last_replayed = 0;

    while (1) {
        if (start_replay == 0)
                break;

        if (!(d_bh = bread (s->s_dev, start_replay, s->s_blocksize))){
            start_replay = next_desc (s, start_replay);
            continue;
        }

        if (!desc_block (d_bh)) {
/*	        printf ("No desc block found at the expected place %lu\n", d_bh->b_blocknr);*/
            brelse (d_bh);
            start_replay = next_desc (s, start_replay);
            continue;
        }

        desc = bh_desc (d_bh);

        if (!is_valid_transaction (s, d_bh)) {
/*	       printf ("skip invalid transaction %ld (length %ld) starting from %lu\n", desc->j_trans_id, desc->j_len, d_bh->b_blocknr);*/
            brelse (d_bh);
            start_replay = next_desc (s, start_replay);
            continue;
        }
	
        if (desc->j_trans_id <= last_replayed) {
            /* we found transaction that has been replayed already */
            brelse (d_bh);
/*	       printf ("Found transaction %ld. last replayed %d\n", desc->j_trans_id, last_replayed);*/
            break;
        }
/*  printf ("Replay transaction %ld (length %ld)-", desc->j_trans_id, desc->j_len);*/


        /* replay transaction */
        {
            int trans_offset = d_bh->b_blocknr - first_journal_block;
            struct buffer_head * log_bh, * in_place;

            block = first_journal_block + ((trans_offset + desc->j_len + 1) % JOURNAL_BLOCK_COUNT);
            while(!(c_bh = bread (s->s_dev, block, s->s_blocksize))){
                if (++block == first_journal_block + journal_size)
                    block = first_journal_block;
                if (block == first_journal_block + ((trans_offset + desc->j_len + 1) % JOURNAL_BLOCK_COUNT))
                    reiserfs_panic(s, "replay_all: too many bad blocks\n");
            }
	
            desc = bh_desc (d_bh);
            commit = bh_commit (c_bh);
            if (does_desc_match_commit (s, desc, commit))
              //	    if (journal_compare_desc_commit(s, desc, commit))
                die ("read_journal: invalid transaction");

            for (i = 0; i < desc->j_len; i ++) {
                /* read from log record */
                while(i < desc->j_len){
                    if (log_bh = bread (s->s_dev, first_journal_block + (trans_offset + 1 + i) % JOURNAL_BLOCK_COUNT,
                                        s->s_blocksize))
                        break;
                    i++;
                }
                if (i == desc->j_len)
                    break;

                if (log_bh->b_blocknr == 8199)
                    printf ("block 8199 put in-placen\n");
                /* write in-place */
                if (i < JOURNAL_TRANS_HALF) {
                     in_place = getblk(s->s_dev, desc->j_realblock[i], s->s_blocksize) ;
                } else {
                    in_place = getblk(s->s_dev, commit->j_realblock[i - JOURNAL_TRANS_HALF], s->s_blocksize) ;
                }
                if (log_bh->b_blocknr == 8199)
                    printf ("Put 8199 to %lu\n", in_place->b_blocknr);
                memcpy (in_place->b_data, log_bh->b_data, s->s_blocksize);
                mark_buffer_dirty (in_place, 0);
                mark_buffer_uptodate (in_place, 1);
                bwrite (in_place);//return -1 if bad block
                brelse (in_place);
                brelse (log_bh);
            }
            brelse (c_bh);
        }
        valid_transactions --;
        printf ("\b\b\b\b    \b\b\b\b%4d", valid_transactions); fflush (stdout);
        last_replayed = desc->j_trans_id;
        start_replay = next_expected_desc (s, d_bh);
        brelse (d_bh);
    }
    printf (" left .. ok\n");
}


/* the simplest scanning for free block., This should be rare */
static __u32 alloc_block (void)
{
    int i, j;
    int bits = g_sb.s_blocksize * 8;
    int start = SB_JOURNAL_BLOCK (&g_sb) + rs_journal_size (g_sb.u.reiserfs_sb.s_rs) + 1;

    for (i = 0; i < SB_BMAP_NR (&g_sb); i ++) {
	j = find_next_zero_bit (g_new_bitmap[i], bits, start);
	if (j < bits) {
	    mark_block_used (j + i * bits);
	    return j + i * bits;
	}
	start = 0;
    }
    die ("allocate_block: no free blocks");
    return 0;
	
}

static struct buffer_head * copy_contents (struct buffer_head * from)
{
    struct buffer_head * bh;
    __u32 new;

    do{
        new = alloc_block ();
        bh = getblk (from->b_dev, new, from->b_size);
    }while(f_read(bh) == 0);
    memcpy (bh->b_data, from->b_data, bh->b_size);
    mark_buffer_uptodate (bh, 1);
    mark_buffer_dirty (bh, 1);
    bwrite (bh);
    return bh;
}


static int block_from_journal (struct super_block * s, __u32 block)
{
    if (block && block < SB_JOURNAL_BLOCK (s)) {
	printf ("not data block (%d) got into tree. Should not appear, but fixable\n", block);
	return 0;
    }
    if (block >= SB_JOURNAL_BLOCK (s) && block <= SB_JOURNAL_BLOCK (s) + rs_journal_size (s->u.reiserfs_sb.s_rs))
	return 1;
    return 0;
}


// pnum can point to an indirect item or to disk_child's node pointer
static void replace_pointer (struct buffer_head * bh, __u32 * pnum)
{
    struct buffer_head * from, * to;
    
    from = bread (bh->b_dev, le32_to_cpu (*pnum), bh->b_size);
    if (from == 0){
        mark_block_used(le32_to_cpu (*pnum));
        *pnum = 0;
    }else{
        to = copy_contents (from);
        *pnum = cpu_to_le32 (to->b_blocknr);
        brelse (to);
    }
    mark_buffer_dirty (bh, 1);
    brelse (from);
}

/* sometimes indirect items point to blocks from journal. Replace them
   with data blocks. It should be rare case */
static void correct_indirect_items (struct super_block * s, struct buffer_head * bh)
{
    int i, j;
    struct item_head * ih;
    __u32 * unfm;

    ih = B_N_PITEM_HEAD (bh, 0);
    for (i = 0; i < B_NR_ITEMS (bh); i ++, ih ++) {
	if (!is_indirect_ih(ih))
	    continue;
	unfm = (__u32 *)B_I_PITEM (bh, ih);
	for (j = 0; j < I_UNFM_NUM (ih); j++)
	    if (block_from_journal (s, le32_to_cpu (unfm[j])))
		replace_pointer (bh, &unfm[j]);
    }
}


// if this *pbh->b_blocks is from journal - allocate new block, copy contents and 
int release_journal_blocks (struct super_block * s, struct buffer_head ** path, int h)
{
    int i;
    unsigned long block;
    struct buffer_head * bh;
    struct buffer_head ** pbh = &path[h];

    if (block_from_journal (s, (*pbh)->b_blocknr)) {
	// possible for root block only
	if (rs_root_block (s->u.reiserfs_sb.s_rs) != (*pbh)->b_blocknr)
	    die ("release_journal_block: journal block in the tree");
	bh = copy_contents (*pbh);
	/* change super block's pointer to the tree root */
	set_root_block (s->u.reiserfs_sb.s_rs, bh->b_blocknr);
	*pbh = bh;
    }

    if (is_leaf_node (*pbh))
	correct_indirect_items (s, *pbh);
    else {
	for (i = 0; i <= node_item_number (*pbh); i ++)
	    if (block_from_journal (s, child_block_number (*pbh, i)))
		replace_pointer (*pbh, &B_N_CHILD_NUM(*pbh, i));
    }
   return 0;
}































