/*
 *  Copyright 2000 Hans Reiser, see reiserfs/README for licensing and copyright details
 */


#include "include.h"



static int is_correct_leaf (char * buf, int blocksize)
{
    struct block_head * blkh;
    struct item_head * ih;
    int used_space;
    int prev_location;
    int i;
    int nr;

    blkh = (struct block_head *)buf;
    if (!is_leaf_block_head (buf))
	return 0;

    nr = le16_to_cpu (blkh->blk_nr_item);
    if (nr < 1 || nr > ((blocksize - BLKH_SIZE) / (IH_SIZE + MIN_ITEM_LEN)))
	/* item number is too big or too small */
	return 0;

    ih = (struct item_head *)(buf + BLKH_SIZE) + nr - 1;
    used_space = BLKH_SIZE + IH_SIZE * nr + (blocksize - ih_location (ih));
    if (used_space != blocksize - le16_to_cpu (blkh->blk_free_space))
	/* free space does not match to calculated amount of use space */
	return 0;

    // FIXME: it is_leaf will hit performance too much - we may have
    // return 1 here

    /* check tables of item heads */
    ih = (struct item_head *)(buf + BLKH_SIZE);
    prev_location = blocksize;
    for (i = 0; i < nr; i ++, ih ++) {
	if (ih_location (ih) >= blocksize || ih_location (ih) < IH_SIZE * nr)
	    return 0;
	if (ih_item_len (ih) < 1 || ih_item_len (ih) > MAX_ITEM_LEN (blocksize))
	    return 0;
	if (prev_location - ih_location (ih) != ih_item_len (ih))
	    return 0;
	prev_location = ih_location (ih);
    }

    // one may imagine much more checks
    return 1;
}


/* returns 1 if buf looks like an internal node, 0 otherwise */
static int is_correct_internal (char * buf, int blocksize)
{
    struct block_head * blkh;
    int nr;
    int used_space;

    blkh = (struct block_head *)buf;

    if (!is_internal_block_head (buf))
	return 0;
    
    nr = le16_to_cpu (blkh->blk_nr_item);
    if (nr > (blocksize - BLKH_SIZE - DC_SIZE) / (KEY_SIZE + DC_SIZE))
	/* for internal which is not root we might check min number of keys */
	return 0;

    used_space = BLKH_SIZE + KEY_SIZE * nr + DC_SIZE * (nr + 1);
    if (used_space != blocksize - le16_to_cpu (blkh->blk_free_space))
	return 0;

    // one may imagine much more checks
    return 1;
}


// make sure that bh contains formatted node of reiserfs tree of
// 'level'-th level
int is_tree_node (struct buffer_head * bh, int level)
{
    if (B_LEVEL (bh) != level)
	return 0;
    if (is_leaf_node (bh))
	return is_correct_leaf (bh->b_data, bh->b_size);

    return is_correct_internal (bh->b_data, bh->b_size);
}


static int is_desc_block (struct reiserfs_journal_desc * desc)
{
    if (!memcmp(desc->j_magic, JOURNAL_DESC_MAGIC, 8))
	return 1;
    return 0;
}


/* returns code of reiserfs metadata block (leaf, internal, super
   block, journal descriptor), unformatted */
int who_is_this (char * buf, int blocksize)
{
    struct reiserfs_journal_desc * desc;

    if (is_correct_leaf (buf, blocksize))
	return THE_LEAF;

    if (is_correct_internal (buf, blocksize))
	return THE_INTERNAL;

    /* super block? */
    if (is_reiser2fs_magic_string ((void *)buf) || 
	is_reiserfs_magic_string ((void *)buf))
	return THE_SUPER;

    /* journal descriptor block? */
    if (is_desc_block ((void *)buf))
	return THE_JDESC;

    /* contents of buf does not look like reiserfs metadata. Bitmaps
       are possible here */
    return THE_UNKNOWN;
}


/* is this block bitmap block or block from journal or skipped area or
   super block? This works for both journal format only yet */
int not_data_block (struct super_block * s, unsigned long block)
{
    int i;

    if (block < SB_JOURNAL_BLOCK (s) + JOURNAL_BLOCK_COUNT + 1)
    {
	return 1;
    }
    for (i = 0; i < SB_BMAP_NR (s); i ++)
	if (block == SB_AP_BITMAP (s)[i]->b_blocknr)
	{
	    return 1;
	}
    return 0;
}

// in reiserfs version 0 (undistributed bitmap)
//
int get_journal_old_start_must (struct reiserfs_super_block * rs)
{
    return 3 + rs_bmap_nr (rs);
}


// in reiserfs version 1 (distributed bitmap) journal starts at 18-th
//
int get_journal_start_must (int blocksize)
{
    return (REISERFS_DISK_OFFSET_IN_BYTES / blocksize) + 2;
}

//
// 4k only now ! 
//
int uread_super_block (struct super_block * s)
{
    struct buffer_head * bh;
    struct reiserfs_super_block * rs;


    bh = bread (s->s_dev, (REISERFS_DISK_OFFSET_IN_BYTES / 4096), 4096);
    if (!bh)
        reiserfs_warning("uread_super_block: bad block is found at a new superblock location\n");
    else{
        rs = (struct reiserfs_super_block *)bh->b_data;
 
        if (is_reiser2fs_magic_string (rs))
            if (rs_journal_start (rs) == get_journal_start_must (rs_blocksize (rs)))
                // 3.6 or >=3.5.22
                goto found;

        if (is_reiserfs_magic_string (rs))
            if (rs_journal_start (rs) == get_journal_start_must (rs_blocksize (rs)))
                // >3.5.9(10) and <=3.5.21
                goto found;

        /* reiserfs signature is not found at a new location */
        brelse (bh);
    }

    bh = bread (s->s_dev, 2, 4096);
    if (!bh)
        die("uread_super_block: bad block is found at an old superblock location\n");

    rs = (struct reiserfs_super_block *)bh->b_data;

    if (is_reiser2fs_magic_string (rs)) {
        if (rs_journal_start (rs) == get_journal_old_start_must (rs))
            // <3.5.9(10) converted to new format
            goto found;
    }

    if (is_reiserfs_magic_string (rs)) {
        if (rs_journal_start (rs) == get_journal_old_start_must (rs))
            // <3.5.9(10)
            goto found;
    }

    brelse (bh);

 not_found:
    printf ("uread_super_block: neither new nor old reiserfs format found on dev 0x%x\n",
        s->s_dev);
    return 1;

 found:

    s->s_blocksize = le16_to_cpu (rs->s_v1.s_blocksize);

    SB_BUFFER_WITH_SB (s) = bh;
    s->u.reiserfs_sb.s_rs = rs;
    //s->s_op = &reiserfs_sops;
    return 0;
}


static int new_format (struct super_block * s)
{
    return (SB_JOURNAL_BLOCK (s) == get_journal_start_must (s->s_blocksize));
}


int uread_bitmaps (struct super_block * s)
{
    int i, bmp ;
    int error = 0;
    //struct reiserfs_super_block * rs = SB_DISK_SUPER_BLOCK(s);

    SB_AP_BITMAP (s) = getmem (sizeof (struct buffer_head *) * SB_BMAP_NR (s));
    if (!SB_AP_BITMAP (s)) {
        printf ("read_bitmaps: malloc failed\n");
        return 1;
    }

    bmp = SB_BUFFER_WITH_SB (s)->b_blocknr + 1;

    for (i = 0; i < SB_BMAP_NR (s); i ++) {
        SB_AP_BITMAP (s)[i] = bread (s->s_dev, bmp, s->s_blocksize);
        if (!SB_AP_BITMAP (s)[i]) {
            printf ("read_bitmaps: bread #%d bitmap failed (%d is a bad block)\n", i, bmp);
            error = IO_ERROR;
        }
        if (new_format (s))
            bmp = (i + 1) * (s->s_blocksize * 8);
        else
            bmp ++;
    }
    return error;
}


